@@ -4,7 +4,7 @@ use warnings;
 
 use Scalar::Util qw/weaken/;
 use Carp qw/confess croak longmess/;
-use Test::Stream::Util qw/get_tid/;
+use Test::Stream::Util qw/get_tid USE_XS/;
 
 use Test::Stream::Sync;
 use Test::Stream::DebugInfo;
@@ -19,16 +19,14 @@ my %LOADED = (
 );
 
 # Stack is ok to cache.
-my $STACK = Test::Stream::Sync->stack;
-
-my @ON_INIT;
-my @ON_RELEASE;
+our $STACK = Test::Stream::Sync->stack;
+our @ON_INIT;
+our @ON_RELEASE;
+our %CONTEXTS;
 
 sub ON_INIT    { shift; push @ON_INIT => @_ }
 sub ON_RELEASE { shift; push @ON_INIT => @_ }
 
-my %CONTEXTS;
-# Be kind 'rewind'
 sub END {
     my $real = $?;
     my $new  = $real;
@@ -45,7 +43,7 @@ sub END {
 }
 
 use Test::Stream::Exporter qw/import export_to exports export/;
-exports qw/context/;
+export context => USE_XS('1.302004', 'context_xs') || \&context_pp;
 export release => sub($;@) {
     $_[0]->release;
     shift; # Remove undef that used to be our $self reference.
@@ -54,7 +52,7 @@ export release => sub($;@) {
 no Test::Stream::Exporter;
 
 use Test::Stream::HashBase(
-    accessors => [qw/stack hub debug _on_release _depth _err _no_destroy_warning _no_destroy/],
+    accessors => [qw/stack hub debug _on_release _depth _err _no_destroy_warning/],
 );
 
 sub init {
@@ -67,40 +65,42 @@ sub init {
 
 sub snapshot { bless {%{$_[0]}}, __PACKAGE__ }
 
-sub release {
+{
+    no warnings 'once';
+    *release = USE_XS('1.302004', 'release_xs') || \&release_pp;
+    *context = USE_XS('1.302004', 'context_xs') || \&context_pp;
+}
+
+sub release_pp {
     my ($self) = @_;
-    my $cbk  = $self->{+_ON_RELEASE};
-    my $dbg  = $self->{+DEBUG};
-    my $hub  = $self->{+HUB};
-    my $hid  = $hub->hid;
-    my $hcbk = $hub->{_context_release};
+    return $_[0] = undef if Internals::SvREFCNT(%$self) != 2;
 
-    my $snap = $cbk || $hcbk ? $self->snapshot : undef;
+    my $hub = $self->{+HUB};
+    my $hid = $hub->{hid};
 
-    weaken($self);
-    $self->{+_NO_DESTROY} = 1;
-    $_[0] = undef; # Kill this reference
-    delete $self->{+_NO_DESTROY} if $self;
-
-    # Removing this reference did not remove the context, so we do not run our
-    # release hooks yet.
-    return if $CONTEXTS{$hid};
+    if ($self != $CONTEXTS{$hid}) {
+        $_[0] = undef;
+        croak "release() should not be called on a non-canonical context.";
+    }
 
+    # Remove the weak reference, this will also prevent the destructor from
+    # having an issue.
     # Remove the key itself to avoid a slow memory leak
     delete $CONTEXTS{$hid};
-    if ($cbk) {
-        $_->($snap) for reverse @$cbk;
+
+    if (my $cbk = $self->{+_ON_RELEASE}) {
+        $_->($self) for reverse @$cbk;
     }
-    if ($hcbk) {
-        $_->($snap) for reverse @$hcbk;
+    if (my $hcbk = $hub->{_context_release}) {
+        $_->($self) for reverse @$hcbk;
     }
-    $_->($snap) for reverse @ON_RELEASE;
+    $_->($self) for reverse @ON_RELEASE;
+
     return;
 }
 
 sub DESTROY {
     my ($self) = @_;
-    return if $self->{+_NO_DESTROY};
 
     my $hid = $self->{+HUB}->hid;
 
@@ -135,18 +135,24 @@ Trace: $mess
     return;
 }
 
-sub context {
+sub context_pp {
     my %params = (level => 0, wrapped => 0, @_);
 
     croak "context() called, but return value is ignored"
         unless defined wantarray;
 
     my $stack = $params{stack} || $STACK;
-    my $hub = $params{hub} || $stack->top;
+    my $hub = $params{hub} || @$stack ? $stack->[-1] : $stack->top;
     my $hid = $hub->{hid};
     my $current = $CONTEXTS{$hid};
 
     my $level = 1 + $params{level};
+    my ($pkg, $file, $line, $sub) = caller($level);
+    unless ($pkg) {
+        confess "Could not find context at depth $level" unless $params{fudge};
+        ($pkg, $file, $line, $sub) = caller(--$level) while ($level >= 0 && !$pkg);
+    }
+
     my $depth = $level;
     $depth++ while caller($depth + 1) && (!$current || $depth <= $current->{+_DEPTH} + $params{wrapped});
     $depth -= $params{wrapped};
@@ -158,48 +164,15 @@ sub context {
 
     return $current if $current && $current->{+_DEPTH} < $depth;
 
-    my ($pkg, $file, $line, $sub) = caller($level);
-    confess "Could not find context at depth $level"
-        unless $pkg;
-
     # Handle error condition of bad level
-    if ($current) {
-        my $oldframe = $current->{+DEBUG}->frame;
-        my $olddepth = $current->{+_DEPTH};
-
-        my $mess = longmess();
-
-        warn <<"        EOT";
-context() was called to retrieve an existing context, however the existing
-context was created in a stack frame at the same, or deeper level. This usually
-means that a tool failed to release the context when it was finished.
-
-Old context details:
-   File: $oldframe->[1]
-   Line: $oldframe->[2]
-   Tool: $oldframe->[3]
-  Depth: $olddepth
-
-New context details:
-   File: $file
-   Line: $line
-   Tool: $sub
-  Depth: $depth
-
-Trace: $mess
-
-Removing the old context and creating a new one...
-        EOT
-
-        delete $CONTEXTS{$hid};
-        $current->release;
-    }
+    $current->_depth_error([$pkg, $file, $line, $sub, $depth])
+        if $current;
 
     my $dbg = bless(
         {
             frame => [$pkg, $file, $line, $sub],
             pid   => $$,
-            tid   => get_tid(),
+            tid   => $INC{'threads.pm'} ? threads->tid() : 0,
             $hub->debug_todo,
         },
         'Test::Stream::DebugInfo'
@@ -221,15 +194,52 @@ Removing the old context and creating a new one...
 
     $_->($current) for @ON_INIT;
 
-    $params{on_init}->($current) if $params{on_init};
-
     if (my $hcbk = $hub->{_context_init}) {
         $_->($current) for @$hcbk;
     }
 
+    $params{on_init}->($current) if $params{on_init};
+
     return $current;
 }
 
+sub _depth_error {
+    my $self = shift;
+    my ($details) = @_;
+    my ($pkg, $file, $line, $sub, $depth) = @$details;
+
+    my $oldframe = $self->{+DEBUG}->frame;
+    my $olddepth = $self->{+_DEPTH};
+
+    my $mess = longmess();
+
+    warn <<"    EOT";
+context() was called to retrieve an existing context, however the existing
+context was created in a stack frame at the same, or deeper level. This usually
+means that a tool failed to release the context when it was finished.
+
+Old context details:
+   File: $oldframe->[1]
+   Line: $oldframe->[2]
+   Tool: $oldframe->[3]
+  Depth: $olddepth
+
+New context details:
+   File: $file
+   Line: $line
+   Tool: $sub
+  Depth: $depth
+
+Trace: $mess
+
+Removing the old context and creating a new one...
+    EOT
+
+    my $hid = $self->{+HUB}->hid;
+    delete $CONTEXTS{$hid};
+    $self->release;
+}
+
 sub throw {
     my ($self, $msg) = @_;
     $_[0]->release; # We have to act on $_[0] because it is aliased
@@ -273,20 +283,21 @@ sub ok {
     my $self = shift;
     my ($pass, $name, $diag) = @_;
 
-    my $e = Test::Stream::Event::Ok->new(
-        debug => $self->{+DEBUG}->snapshot,
+    my $e = bless {
+        debug => bless( {%{$self->{+DEBUG}}}, 'Test::Stream::DebugInfo'),
         pass  => $pass,
         name  => $name,
-    );
+    }, 'Test::Stream::Event::Ok';
+    $e->init;
 
-    return $self->hub->send($e) if $pass;
+    return $self->{+HUB}->send($e) if $pass;
 
     $diag ||= [];
     unshift @$diag => $e->default_diag;
 
     $e->set_diag($diag);
 
-    $self->hub->send($e);
+    $self->{+HUB}->send($e);
 }
 
 sub note {
